/*
 * Copyright (c) Linux community.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gst/gst.h>
#include <gst/interfaces/propertyprobe.h>

#include "audio-sources.h"
#include "pulseaudio.h"
#include "support.h"  // _(x)
#include "gconf.h"
#include "utility.h"
#include "log.h"
#include "dbus-player.h"

// Note: The get_audio_devices() function has two variants.
// This function returns the list of available audio-devices (audio sources) such as sound-cards, microphones and webcams.
// The 1.st version uses PulseAudio to get the device list.
// The second variant uses GStreamer, but GStreamer fails to return the device name (it is always NULL, at least in Ubuntu 10.x).
// Later we should avoid the PulseAudio dependency and use GStreamer for all functions.

// A List of audio devices (global to this module)
// Contains:
//  * Real audio hardware, such as audio cards, microphones, webcams.
//  * Media Players like RhythmBox, Amarok and Banshee.
//  * Add also Skype if it's installed. Media Players and Skype can automatically control the recording via DBus.

G_LOCK_DEFINE_STATIC(g_device_list);
static GList *g_device_list = NULL;

static void audio_sources_load_device_list();
static void get_audio_devices();

static void audio_sources_get_GNOME_default(gchar **audio_source, gchar **device_id);
static void audio_sources_get_device_for_media_players(gchar **audio_source, gchar **device_id);

void audio_sources_init() {
    LOG_DEBUG("Init audio-sources.c.\n");

    g_device_list = NULL;

    // Init Puleaudio module
    pulseaudio_module_init();

    // Init DBus/Media Player modules
    dbus_player_init();
}

void audio_sources_exit() {
    LOG_DEBUG("Clean up audio-sources.c.\n");

    // Clean up pulseaudio molule
    pulseaudio_module_exit();

    // Clean up media players/DBus
    dbus_player_exit();

    audio_sources_free_list(g_device_list);
    g_device_list = NULL;
}

// ---------------------------------------------------
// DeviceItem functions

DeviceItem *device_item_create(gchar *id, gchar *description) {
    DeviceItem *item = g_malloc0(sizeof(DeviceItem));
    if (id) item->id = g_strdup(id);
    if (description) item->description = g_strdup(description);
    item->sink_index = PA_INVALID_INDEX;
    return item;
}

DeviceItem *device_item_copy(DeviceItem *item) {
    // Make a copy of DeviceItem
    DeviceItem *copy = (DeviceItem*)g_memdup(item, sizeof(DeviceItem));
    copy->id = g_strdup(item->id);
    copy->description = g_strdup(item->description);
    return copy;
}

void device_item_free(DeviceItem *item) {
    if (!item) return;
    // Free the values
    if (item->id) g_free(item->id);
    if (item->description) g_free(item->description);
    // Free the node
    g_free(item);
}

const gchar *device_item_get_type_name(guint type) {
    switch (type) {
    case NOT_DEFINED:
        return "NOT DEFINED";

    case DEFAULT_DEVICE:
        return "DEFAULT_DEVICE";

    case AUDIO_SINK:
        return "AUDIO_SINK";

    case AUDIO_RECEIVER:
        return "AUDIO_RECEIVER";

    case AUDIO_SOURCE:
        return "AUDIO_SOURCE";

    case MEDIA_PLAYER:
        return "MEDIA_PLAYER";
    }
    return "UNKNOWN TYPE";
}

// ---------------------------------------------------
// These are filter-functions used by audio_sources_get_with_filter(...).

gboolean filter_for_sink_dev(DeviceItem *item) {
    if (!item) return FALSE;

    // Accept device if it is AUDIO_SOURCE and it is related to a real audio sink/audio card.
    if (item->type == AUDIO_SOURCE && item->sink_index != PA_INVALID_INDEX) {
        return TRUE;
    }
    return FALSE;
}

// ---------------------------------------------------
// Functions related to audio sources

void audio_sources_free_list(GList *lst) {
    // Free the list and all its DeviceItems
    g_list_foreach(lst, (GFunc)device_item_free, NULL);
    g_list_free(lst);
    lst = NULL;
}

void audio_sources_print_list_ex() {
    // Print device list
    audio_sources_print_list(g_device_list, "DEVICE LIST");
}

void audio_sources_print_list(GList *list, gchar *tag) {
    // Print device list
    LOG_MSG("\n%s:\n", tag);

    GList *l = g_list_first(list);
    gint i = 0;
    while (l) {
        DeviceItem *item = (DeviceItem*)l->data;

        const gchar *type_name = device_item_get_type_name(item->type);

        LOG_MSG("#%d:type=%s(%d) %s, id=%s  sink_index=%d\n", i++, type_name, item->type, item->description, item->id, item->sink_index);
        l = g_list_next(l);
    }
    LOG_MSG("-------------------------------------------\n");
}

DeviceItem *audio_sources_find_id(gchar *device_id) {
    // Return DeviceItem for given device_id.

    // Lock
    G_LOCK(g_device_list);

    // Let's assume g_device_list already loaded.
    GList *l = g_list_first(g_device_list);
    DeviceItem *found_item = NULL;
    while (l) {
        DeviceItem *item = (DeviceItem*)l->data;
        // Compare ids
        if (!g_strcmp0(item->id, device_id) || (device_id == NULL && item->id == NULL)) {
            found_item = item;
            break;
        }
        l = g_list_next(l);
    }

    // Unlock
    G_UNLOCK(g_device_list);

    return found_item;
}

GList *audio_sources_get_for_type(gint type) {
    // Return a GList of DeviceItems that match the type. The type can be a OR'ed value.
    // Eg. GList *copy = audio_sources_get_for_type(AUDIO_SOURCE | DEFAULT_DEVICE);
    //     ...
    //     audio_sources_free_list(copy);
    //     copy = NULL;

    // Set lock
    G_LOCK(g_device_list);

    GList *lst = NULL;

    GList *n = g_list_first(g_device_list);
    while (n) {
        DeviceItem *item = (DeviceItem*)n->data;

        // Test the type
        if (item->type & type) {
            DeviceItem *copy = device_item_copy(item);
            lst = g_list_append(lst, copy);
        }

        // Next item
        n = g_list_next(n);
    }

    // Free the lock
    G_UNLOCK(g_device_list);

    // Notice:
    // The returned list is a copy. You must free it with
    // audio_sources_free_list(lst);
    // lst = NULL;
    return lst;
}

GList *audio_sources_get_with_filter(DeviceItemFilterFunc filter_func) {
    // Return a GList of DeviceItems that are filtered/accepted by filter_func.
    // Eg. GList *copy = audio_sources_get_with_filter(some_filter_func);
    //     ...
    //     audio_sources_free_list(copy);
    //     copy = NULL;

    // Lock
    G_LOCK(g_device_list);

    GList *lst = NULL;

    GList *n = g_list_first(g_device_list);
    while (n) {
        DeviceItem *item = (DeviceItem*)n->data;

        if (filter_func && filter_func(item)) {
            DeviceItem *copy = device_item_copy(item);
            lst = g_list_append(lst, copy);
        }

        // Next item
        n = g_list_next(n);
    }

    // Unlock
    G_UNLOCK(g_device_list);

    // Notice:
    // The returned list is a copy. You must free it with
    // audio_sources_free_list(lst);
    // lst = NULL;
    return lst;
}

void audio_sources_device_changed(gchar *device_id) {
    // Device changed.

    // Disconnect/re-connect DBus signals (if the device_id is Media Player or Skype)
    dbus_player_player_changed(device_id/*player->service_name*/);
}

void audio_sources_get_device(gchar **audio_source, gchar **device_id) {
    // Return audio_source and device_id as set in the GUI.
    // Handle both MEDIA_PLAYERs, Skype and "default-device".
    *audio_source = NULL;
    *device_id = NULL;

    gchar *dev = NULL;
    gint type = -1;

    // Lock
    G_LOCK(g_device_list);

    conf_get_string_value("audio_device_id", &dev);
    conf_get_int_value("audio_device_type", &type);

    gchar *source = g_strdup(DEFAULT_AUDIO_SOURCE);

    // Record from a Media Player, Skype?
    if (type == MEDIA_PLAYER) {
        g_free(dev);
        g_free(source);

        // Get real output device (sink device) for Media Players, Skype?
        // This is set in the [Additional settings] dialog.
        audio_sources_get_device_for_media_players(&source, &dev);

        // Set to System's default device"?
    } else if (!g_strcmp0(dev, "default-device")) {
        g_free(dev);
        g_free(source);
        audio_sources_get_GNOME_default(&source, &dev);
    }

    // Unlock
    G_UNLOCK(g_device_list);

    // Return values
    *audio_source = source;
    *device_id = dev;

    // Caller should g_free() both audio_source and device_id.
}

static void audio_sources_get_GNOME_default(gchar **audio_source, gchar **device_id) {
    // TODO: What is default device?

    // So what is this "Default" device/audio-source then?
    // Ref: http://ubuntuforums.org/showpost.php?p=9686252&postcount=15

    // Try first "gconfaudiosrc"
    // Its real value is probably set in the GConf registry.
    // Start gconf-editor and browse to key /system/gstreamer/0.10/default (audiosrc).
    gchar *source = g_strdup("gconfaudiosrc");
    GstElement *elem = gst_element_factory_make(source, "test-audio-source");
    gboolean ret = (elem != NULL);
    g_object_unref(elem);

    if (!ret) {
        g_free(source);

        // Try "autoaudiosrc"
        source = g_strdup("autoaudiosrc");
        elem = gst_element_factory_make(source, "test-audio-source");
        ret = (elem != NULL);
        g_object_unref(elem);
    }

    if (!ret) {
        g_free(source);
        source = g_strdup(DEFAULT_AUDIO_SOURCE);
    }

    // Return values
    *audio_source = source;

    // Let audio_source settle the device.
    *device_id = NULL;

    // Caller should g_free() both audio_source and device_id.
}

static void audio_sources_get_device_for_media_players(gchar **audio_source, gchar **device_id) {
    // Get the value from GConf (this is set in the [Additional settings] dialog)

    *audio_source = NULL;
    *device_id = NULL;

    gchar *dev = NULL;
    conf_get_string_value("device_for_media_players", &dev);

    gchar *source = g_strdup(DEFAULT_AUDIO_SOURCE);

    // Set to System's default device" or empty?
    if (!g_strcmp0(dev, "default-device")) {
        g_free(dev);
        g_free(source);
        audio_sources_get_GNOME_default(&source, &dev);

    } else if (str_length(dev, 2048) < 1) {
        g_free(dev);
        dev = NULL;

        // Take the first id that is related to a real audio card (sink_index >= 0)
        GList *n = g_list_first(g_device_list);
        while (n) {
            DeviceItem *item = (DeviceItem*)n->data;
            if (item->type == AUDIO_SOURCE && item->sink_index != PA_INVALID_INDEX) {
                dev = g_strdup(item->id);
                break;
            }
            n = g_list_next(n);
        }
    }

    *audio_source = source;
    *device_id = dev;

    // Caller should g_free() both audio_source and device_id.
}

// ---------------------------------------------------

static void audio_sources_load_device_list() {
    // Lock
    G_LOCK(g_device_list);

    // Free the old list
    audio_sources_free_list(g_device_list);
    g_device_list = NULL;

    // 1) Get list of real audio devices (fill to g_device_list)
    get_audio_devices();

    // 2) Add Media Players, Skype etc. to g_device_list (these can control the recording via DBus)
    // See the dbus_xxxx.c modules
    // ----------------------------

    // Get the player list (it's a GHashTable)
    GHashTable *player_list = dbus_player_get_player_list();
    GHashTableIter hash_iter;
    gpointer key, value;

    // Add to the g_device_list
    g_hash_table_iter_init(&hash_iter, player_list);
    while (g_hash_table_iter_next(&hash_iter, &key, &value)) {
        MediaPlayerRec *p = (MediaPlayerRec*)value;

        gchar *t = p->app_name;
        if (!t)
            t = p->service_name;

        // Eg. "Record from Banshee Meda Player" or "Record from Skype"
        gchar *txt = g_strdup_printf(_("Record from %s"), t);
        DeviceItem *item = device_item_create(p->service_name, txt);

        // Important: set type to MEDIA_PLAYER
        item->type = MEDIA_PLAYER;

        g_device_list = g_list_append(g_device_list, item);
        g_free(txt);
    }

    // The g_device_list is now loaded and ready

    // Unlock
    G_UNLOCK(g_device_list);
}

GtkWidget *audio_sources_create_combo() {
    // Create a GtkComboBox with N_DEVICE_COLUMNS

    // Create list store
    GtkListStore *store = gtk_list_store_new(N_DEVICE_COLUMNS, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING);

    GtkWidget *combo = gtk_combo_box_new();
    gtk_combo_box_set_model(GTK_COMBO_BOX(combo),GTK_TREE_MODEL(store));

    // Unref store
    g_object_unref(G_OBJECT(store));

    // Device type (see the description of DeviceType enum)
    GtkCellRenderer *cell = gtk_cell_renderer_text_new();
    g_object_set(cell, "visible", 0, NULL);
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), cell, TRUE);
    gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(combo), cell, "text", COL_DEVICE_TYPE, NULL);

    // Device id column, invisible
    cell = gtk_cell_renderer_text_new();
    g_object_set(cell, "visible", 0, NULL);
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), cell, TRUE);
    gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(combo), cell, "text", COL_DEVICE_ID, NULL);

    // Device description column, visible
    cell = gtk_cell_renderer_text_new();
    gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), cell, TRUE);
    gtk_cell_layout_set_attributes(GTK_CELL_LAYOUT(combo), cell, "text", COL_DEVICE_DESCR, NULL);

    return combo;
}

void audio_source_fill_combo(GtkWidget *combo) {
    GtkTreeModel *model = gtk_combo_box_get_model(GTK_COMBO_BOX(combo));
    GtkTreeIter  iter;

    if (!GTK_IS_LIST_STORE(GTK_LIST_STORE(model))) {
        return;
    }

    // Disable "changed" signal
    gulong signal_id = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(combo), "selection-changed-signal"));
    g_signal_handler_block(combo, signal_id);

    // Clear the old combo list
    gtk_list_store_clear(GTK_LIST_STORE(model));

    // Get a new list of:
    // 1) Audio devices, audio-card w/ loudspeakers, webcams, microphones, etc.
    // 2) + Add media players we can see on the DBus.
    audio_sources_load_device_list();

    // Set lock
    G_LOCK(g_device_list);

    // Then add items to the ComboBox
    GList *n = g_list_first(g_device_list);
    while (n) {
        DeviceItem *item = (DeviceItem*)n->data;

        // Add new row
        gtk_list_store_append(GTK_LIST_STORE(model), &iter);

        // Set column data
        gtk_list_store_set(GTK_LIST_STORE(model), &iter, COL_DEVICE_TYPE, item->type/*type*/,
                           COL_DEVICE_ID, item->id /* Internal device id or media player id */ ,
                           COL_DEVICE_DESCR, item->description /* Visible text */, -1);
        // Next item
        n = g_list_next(n);
    }

    // Enable changed-signal
    g_signal_handler_unblock(combo, signal_id);

    // Free the lock
    G_UNLOCK(g_device_list);
}

void audio_sources_combo_set_id(GtkWidget *combo, gchar *device_id) {
    // Set combo selection by device id

    GtkTreeModel *model = gtk_combo_box_get_model(GTK_COMBO_BOX(combo));
    if (!GTK_IS_TREE_MODEL(model)) {
        return;
    }

    GtkTreeIter  iter;

    gboolean valid = gtk_tree_model_get_iter_first(model, &iter);
    while (valid) {
        gchar *id = NULL;
        gtk_tree_model_get(model, &iter, COL_DEVICE_ID, &id, -1);

        // Compare ids
        if (!g_strcmp0(device_id, id)) {
            // Select it
            gtk_combo_box_set_active_iter(GTK_COMBO_BOX(combo), &iter);
            return;
        }
        valid = gtk_tree_model_iter_next(model, &iter);
    }

    // str_value was not found. Select the first, 0'th row
    valid = gtk_tree_model_get_iter_first(model, &iter);
    if (valid) {
        gtk_combo_box_set_active_iter(GTK_COMBO_BOX(combo), &iter);
    }
}

gboolean audio_sources_combo_get_values(GtkWidget *combo, gchar **device_name, gchar **device_id, gint *device_type) {
    *device_name = NULL;
    *device_id = NULL;
    *device_type = -1;

    GtkTreeModel *model = gtk_combo_box_get_model(GTK_COMBO_BOX(combo));
    if (!GTK_IS_TREE_MODEL(model)) {
        return FALSE;
    }

    GtkTreeIter iter;
    if (!gtk_combo_box_get_active_iter(GTK_COMBO_BOX(combo), &iter)) return FALSE;

    gtk_tree_model_get(model, &iter, COL_DEVICE_ID, device_id,
                       COL_DEVICE_DESCR, device_name,
                       COL_DEVICE_TYPE, device_type, -1);

    // The caller should g_free() both device_name and device_id.
    return TRUE;
}

// Choose one of these get_audio_devices() functions

#if 1
static void get_audio_devices() {
    // Use PulseAudio to get the device ids and names.
    // This returns both device id and correct name/description.
    // Compilation requires -lpulse: pkg-config --cflags --libs libpulse
    GList *pa_list = pa_get_source_list();

    // Take a copy of the PulseAudio's data. Let all modules (puleaudio.c, audio-source.c, etc.) have their own private lists.
    // So this module can erase its list without interfering others' data.

    GList *n = g_list_first(pa_list);
    while (n) {
        DeviceItem *item = (DeviceItem*)n->data;

        // Take a copy
        DeviceItem *copy = device_item_copy(item);

        // Add to our private list
        g_device_list = g_list_append(g_device_list, copy);

        n = g_list_next(n);
    }

    // Add "Default" device
    // See also: http://pulseaudio.org/wiki/DefaultDevice

    // Translators: This is system's default audio device.
    DeviceItem *item = device_item_create("default-device"/*id*/, _("System's default device")/*description*/);
    item->type = DEFAULT_DEVICE;
    g_device_list = g_list_append(g_device_list, item);
}
#endif

#if 0
static void get_audio_devices() {
    // Get audio devices (audio sources) from GStreamer.
    // This returns correct device id, but the name/description is NULL (at least for some audio cards).

    gchar *pipe_name = DEFAULT_AUDIO_SOURCE;

    GstElementFactory *factory;
    GstElement *element;

    factory = gst_element_factory_find(pipe_name);

    if (!factory) {
        GST_WARNING("Failed to find factory for pipeline %s.\n", pipe_name);
        return;
    }

    element = gst_element_factory_create(factory, "test");
    const gchar *longname = gst_element_factory_get_longname(factory);
    if (!element) {
        GST_WARNING("Failed to create instance of factory '%s' (%s)", longname, GST_PLUGIN_FEATURE (factory)->name);
        return;
    }

    // const gchar *klass_name = gst_element_factory_get_klass(factory);
    // const gchar *description = gst_element_factory_get_description(factory);
    // const gchar *author = gst_element_factory_get_author(factory);

    gint n;
    gchar *name;
    GValueArray *array;

    GObjectClass *klass = G_OBJECT_GET_CLASS(element);

    GstPropertyProbe *probe = GST_PROPERTY_PROBE(element);
    const GParamSpec *pspec = gst_property_probe_get_property(probe, "device");

    /* Do we have a "device" property? */
    if (!g_object_class_find_property(klass, "device") || !GST_IS_PROPERTY_PROBE(element) || !probe || !pspec) {
        GST_DEBUG ("Found source '%s' (%s) - no device", longname, GST_PLUGIN_FEATURE (factory)->name);
        return;
    }

    /* Set autoprobe[-fps] to FALSE to avoid delays when probing. */
    if (g_object_class_find_property(klass, "autoprobe")) {
        g_object_set(G_OBJECT(element), "autoprobe", FALSE, NULL);
    }

    if (g_object_class_find_property(klass, "autoprobe-fps")) {
        g_object_set(G_OBJECT (element), "autoprobe-fps", FALSE, NULL);
    }

    array = gst_property_probe_probe_and_get_values(probe, pspec);
    if (array != NULL) {

        /* default device item, so we can let the element handle it */
        if (array->n_values > 0) {
            g_print("This is the default device (default)\n");
        }

        for (n = 0; n < array->n_values; n++) {
            GValue *device;
            device = g_value_array_get_nth(array, n);
            g_object_set_property(G_OBJECT(element), "device", device);

            if (gst_element_set_state(element, GST_STATE_READY) != GST_STATE_CHANGE_SUCCESS) {
                GST_WARNING("Found source '%s' (%s) - device %s failed to open", longname, GST_PLUGIN_FEATURE(factory)->name, g_value_get_string(device));
                continue;
            }

            gchar *device_id = (gchar*)g_value_get_string(device);

            g_object_get(G_OBJECT(element), "device-name", &name, NULL);

            // TODO: Fix this bug. The name is always NULL.
            // See Launchpad bug: https://bugs.launchpad.net/ubuntu/+source/gstreamer0.10/+bug/607162

            // NULL?
            if (!name) {
                // Show id
                name = device_id;
            }

            DeviceItem *item = device_item_create(device_id, name);

            // Is this an "Audio Output" device?
            // TODO: Find a better way to determine if this is a sink (audio output) device.
            if (g_str_has_suffix(device_id, ".monitor")) {
                // Set the type
                item->type = AUDIO_SOURCE;
            }

            g_device_list = g_list_append(g_device_list, item);

            // g_print("M1: Found source for '%s' (%s) - device %s '%s'\n", longname, GST_PLUGIN_FEATURE(factory)->name, g_value_get_string(device), name);
        }
    }

    gst_element_set_state(element, GST_STATE_NULL);
    gst_object_unref(GST_OBJECT(element));
    gst_object_unref(GST_OBJECT(factory));

    // Add "Default" device
    // See also: http://pulseaudio.org/wiki/DefaultDevice
    DeviceItem *item = device_item_create("default-device"/*id*/, _("System's default device")/*description*/);
    item->type = DEFAULT_DEVICE;
    g_device_list = g_list_append(g_device_list, item);
}
#endif

