/*
 * Copyright (c) Linux community.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
*/
#include "gst-listener.h"
#include "audio-sources.h"
#include "gconf.h"
#include "log.h"

#include <string.h>
#include <glib.h>
#include <gdk/gdk.h>
#include <math.h>
#include <gst/gst.h>

#define MAX_LEVEL_VALUES 50 // 10 * 5 = 50, ~5 seconds of data

typedef struct {
    guint sum_count;  // Number of collected values

    double sum_rms_dB;  // Average rms level (dB)
    double sum_rms; // Average rms level (0 - 1.0)
} LevelValues;

// Level values
G_LOCK_DEFINE_STATIC(g_level_values);
static LevelValues g_level_values;

// GStreamer pipeline
static GstElement *g_listener_pipeline;

// Debug flag
static gboolean g_debug_flag = FALSE;

// Current audio source and device id
static gchar *g_curr_audio_source = NULL;
static gchar *g_curr_device_id = NULL;

void listener_module_init();
void listener_module_exit();

static void listener_set_data(GstClockTime end_time, double rms_dB, double peak_dB, double decay_dB);

static gboolean listener_is_listening();
static gboolean listener_create_pipeline();
static void listener_shutdown_pipeline();

void listener_module_init() {
    LOG_DEBUG("Init gst-listener.c.\n");

    // Initialize
    g_listener_pipeline = NULL;
    g_curr_audio_source = NULL;
    g_curr_device_id = NULL;

    // Reset data record
    listener_clear_data();

    g_debug_flag = FALSE;
}

void listener_module_exit() {
    LOG_DEBUG("Clean up gst-listener.c.\n");

    listener_stop_listening();
}

void listener_set_debug_flag(gboolean on_off) {
    // Set debug flag.
    // Listener process will now print out level values so user can study them.
    g_debug_flag = on_off;
}

void listener_start_listening() {
    gchar *audio_source = NULL;
    gchar *device_id = NULL;

    // Get audio source and device id.
    audio_sources_get_device(&audio_source, &device_id);

    gboolean changed = g_strcmp0(g_curr_audio_source, audio_source);
    changed = changed || g_strcmp0(g_curr_device_id, device_id);

    // Device changed?
    if (changed) {
        // Yes, stop the listener.
        listener_stop_listening();
    }

    // Already listening (not changed)?
    if (listener_is_listening()) {
        g_free(audio_source);
        g_free(device_id);
        return;
    }

    // Save new values
    if (g_curr_audio_source)
        g_free(g_curr_audio_source);

    g_curr_audio_source = audio_source;

    if (g_curr_device_id)
        g_free(g_curr_device_id);

    g_curr_device_id = device_id;

    // Start listener.
    // Create GStreamer pipeline and monitor signal levels.
    listener_create_pipeline();
}

void listener_stop_listening() {
    listener_shutdown_pipeline();

    if (g_curr_audio_source)
        g_free(g_curr_audio_source);

    g_curr_audio_source = NULL;

    if (g_curr_device_id)
        g_free(g_curr_device_id);

    g_curr_device_id = NULL;
}

static gboolean listener_is_listening() {
    // The pipeline has been created?
    if (!GST_IS_ELEMENT(g_listener_pipeline)) return FALSE;

    // The state is?
    // Ref: http://www.gstreamer.net/data/doc/gstreamer/head/gstreamer/html/GstElement.html#gst-element-get-state
    GstState status, pending;
    gst_element_get_state(g_listener_pipeline, &status, &pending, 0);

    return (status == GST_STATE_PLAYING);
}

void listener_clear_data() {
    // Lock
    G_LOCK(g_level_values);

    // Zero all values
    g_level_values.sum_rms_dB = 0.0;
    g_level_values.sum_rms = 0.0;
    g_level_values.sum_count = 0;

    // Unlock
    G_UNLOCK(g_level_values);
}

guint listener_get_average_rms(gdouble *avg_rms_dB, gdouble *avg_rms) {
    // Lock
    G_LOCK(g_level_values);

    guint count = 0;
    if (g_level_values.sum_count > 0) {
        // Average rms (dB)
        *avg_rms_dB = g_level_values.sum_rms_dB / g_level_values.sum_count;

        // Average rms (0 - 1.0)
        *avg_rms = g_level_values.sum_rms / g_level_values.sum_count;

        count = g_level_values.sum_count;
    }

    // Convert form 0 - 1.0 to 0 - 100%.
    *avg_rms = *avg_rms * 100.0;

    // Unlock
    G_UNLOCK(g_level_values);

    // Number of values
    return count;
}

static void listener_set_data(GstClockTime end_time, double rms_dB, double peak_dB, double decay_dB) {
    // Save level values.
    // Please study:
    // http://www.gstreamer.net/data/doc/gstreamer/head/gst-plugins-good-plugins/html/gst-plugins-good-plugins-level.html
    // http://en.wikipedia.org/wiki/Decibel

    // Set lock
    G_LOCK(g_level_values);

    if (rms_dB == -INFINITY) {
        ;
    }

    // rms_dB:
    // RMS, http://en.wikipedia.org/wiki/Root_mean_square
    // From dB to a normalized 0 - 1.0 value.
    double rms = pow(10, rms_dB / 20);

    // peak_dB:
    // Signal peak from dB to 0 - 1.0
    double peak = exp(peak_dB / 20);
    if (peak > 1.0) peak = 1.0;

    // decay_dB:
    // Decay rate of signal (in dB/sec), not used.

    // RMS is the only meaningful value.
    if (g_level_values.sum_count >= MAX_LEVEL_VALUES) {
        g_level_values.sum_rms_dB = 0.0;
        g_level_values.sum_rms = 0.0;
        g_level_values.sum_count = 0;
    }

    g_level_values.sum_rms_dB += rms_dB;
    g_level_values.sum_rms  += rms;
    g_level_values.sum_count += 1;

    // Show debug info in a terminal window?
    if (g_debug_flag) {

        // Calculate average
        gdouble avg_rms_dB = g_level_values.sum_rms_dB / g_level_values.sum_count;
        gdouble avg_rms = g_level_values.sum_rms / g_level_values.sum_count;

        LOG_MSG("Time: %" GST_TIME_FORMAT ": rms=%3.1f dB(%3.1f%%) peak=%3.1f dB(%3.1f%%) decay=%3.1f dB. Value count=%d Avg.rms=%3.1f dB(%3.1f%%)\n",
                GST_TIME_ARGS(end_time), rms_dB, rms*100.0, peak_dB, peak*100.0,  decay_dB,
                g_level_values.sum_count, avg_rms_dB, avg_rms*100.0);
    }

    // Unlock
    G_UNLOCK(g_level_values);
}

static gboolean listener_message_handler(GstBus * bus, GstMessage * message, gpointer data) {
    if (message->type == GST_MESSAGE_ELEMENT) {

        const GstStructure *s = gst_message_get_structure (message);
        const gchar *name = gst_structure_get_name (s);

        if (strcmp (name, "level") == 0) {
            GstClockTime endtime;

            if (!gst_structure_get_clock_time (s, "endtime", &endtime)) {
                g_warning("Could not parse endtime");
            }

            // We can get the number of channels as the length of any of the value lists
            const GValue *list;
            list = gst_structure_get_value(s, "rms");

            gint channels = 0;
            channels = gst_value_list_get_size(list);

            gint i = 0;
            for (i = 0; i < channels; ++i) {
                const GValue *value;

                list = gst_structure_get_value(s, "rms");

                value = gst_value_list_get_value(list, i);
                gdouble rms_dB = g_value_get_double(value);

                list = gst_structure_get_value(s, "peak");
                value = gst_value_list_get_value (list, i);
                gdouble peak_dB = g_value_get_double(value);

                list = gst_structure_get_value(s, "decay");
                value = gst_value_list_get_value(list, i);
                gdouble decay_dB = g_value_get_double(value);

                listener_set_data(endtime, rms_dB, peak_dB, decay_dB);

            }
        }
    }

    // We handled the message we want, and ignored the ones we didn't want. so the core can unref the message for us
    return TRUE;
}

static void listener_shutdown_pipeline() {
    // Shutdown the listener pipeline

    if (!GST_IS_PIPELINE(g_listener_pipeline)) {
        goto LBL_1;
    }

    LOG_LISTENER("Shutdown listener pipeline.\n");

    // Switch to GST_STATE_NULL
    gst_element_set_state(g_listener_pipeline, GST_STATE_NULL);

    // Then destroy it
    gst_object_unref(GST_OBJECT(g_listener_pipeline));
    g_listener_pipeline = NULL;

LBL_1:
    listener_clear_data();
}

static gboolean listener_create_pipeline() {

    if (GST_IS_PIPELINE(g_listener_pipeline)) {
        // Shutdown first
        listener_shutdown_pipeline();
    }

    LOG_LISTENER("Creating listener pipeline for %s and %s.\n", g_curr_audio_source, g_curr_device_id);

    GstCaps *caps = gst_caps_from_string("audio/x-raw-int,channels=1");

    g_listener_pipeline = gst_pipeline_new("Listener for the audio-recorder application");
    if (!GST_IS_PIPELINE(g_listener_pipeline)) {
        goto L_END1;
    }

    GstElement *audio_source = gst_element_factory_make(g_curr_audio_source, NULL);
    if (!audio_source) {
        goto L_END1;
    }

    if (g_curr_device_id) {
        // Set the device we'll monitor and listen to
        g_object_set(G_OBJECT(audio_source), "device", g_curr_device_id, NULL);
    }

    GstElement *audio_convert = gst_element_factory_make("audioconvert", NULL);
    if (!audio_convert) {
        goto L_END1;
    }

    GstElement *level = gst_element_factory_make("level", NULL);
    if (!level) {
        goto L_END1;
    }

    GstElement *fake_sink = gst_element_factory_make("fakesink", NULL);
    if (!fake_sink) {
        goto L_END1;
    }

    gst_bin_add_many(GST_BIN(g_listener_pipeline), audio_source, audio_convert, level, fake_sink, NULL);
    g_assert(gst_element_link(audio_source, audio_convert));
    g_assert(gst_element_link_filtered(audio_convert, level, caps));
    g_assert(gst_element_link(level, fake_sink));

    // Receive signal level messages
    g_object_set(G_OBJECT(level), "message", TRUE, NULL);

    // Sync
    g_object_set(G_OBJECT(fake_sink), "sync", TRUE, NULL);

    GstBus *bus = gst_element_get_bus(g_listener_pipeline);
    gst_bus_add_signal_watch(bus);
    g_signal_connect(bus, "message::element", G_CALLBACK(listener_message_handler), NULL);

    // Start GST_STATE_PLAYING
    gst_element_set_state(g_listener_pipeline, GST_STATE_PLAYING);

    gst_object_unref(bus);

    // Ok
    return TRUE;

L_END1:
    // Error occured

    // Shutdown
    listener_shutdown_pipeline();

    // Show error message
    LOG_ERROR("Cannot create listener for %s and %s.\n",  g_curr_audio_source, g_curr_device_id);
    return FALSE;
}

